#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/unistd.h> /* syscall numbers */
#include <linux/ptrace.h> /* struct pt_regs */
#include <linux/sched.h>  /* struct task_struct */
#include <linux/interrupt.h> 

#include <asm/processor.h>
#include <asm/system.h>
#include <asm/current.h>
#include <asm/debug.h>
#include <asm/assert.h>

#include <libdev/etiContext.h>
#include <libdev/etiExceptions.h>

#include "context.h"


typedef asmlinkage int
             (*syscall_t)(
                     void* a0, 
                     void* a1, 
                     void* a2, 
                     void* a3, 
                     void* a4, 
                     void* a5, 
                     void* a6, 
                     ctxSyscallContext_t  ctx
              );


extern syscall_t sys_call_table[];

 
#define MARGIN 1000

void __CHECK_STACK(char *s)
{
#if 1
    u32 c   = (u32)  current;
    u32 cv  = (u32)  current_var;
    u32 end = (u32) (current+1);
    u32 here= (u32) &here;
    
    if (0) {
         static int count;

         if (count-- == 0) { printk("\n"); count= 79; } 
         printk("%s", s);
    }

#if 0
    ASSERT (  c == cv, 
             ("@@@@@%s not on kernel stack!!!\n 0x%08x  0x%08x\n",
                    s, current, current_var) ); 
#endif

    ASSERT ( (here - end) >= MARGIN,
             ("@@@@@%s kernel stack overflow!!!\n 0x%08x  0x%08x\n",
                    s, current, here) ); 

    ASSERT ( (here & 0x80000000) != 0,
             ("@@@@@%s stack pointer not on kernel stack!!!\n 0x%08x  0x%08x\n",
                    s, current, here) ); 
#endif
}    


void syscall_trace(void);

/* C syscall handler  */
int syscall_handler( void* a0, 
                     void* a1, 
                     void* a2, 
                     void* a3, 
                     void* a4, 
                     void* a5, 
                     void* a6, 
                     u32                  function_code,
                     ctxSyscallContext_t  ctx
                     ) 
{
    struct task_struct *tsk = current;

   /* Clear ESR: */ 
    hmpv_write_esr_cl0(hmpv_read_esr_cl0() & ~bspregSYS);

    if (function_code < __NR_Linux) {
	    return -ENOSYS;
    } else
        
    if (function_code > __NR_ETI + __NR_ETI_syscalls) {
	    return -ENOSYS;
    }
    else {
            u32       flags;
            int       retval;
            syscall_t syscall= sys_call_table[function_code];
	    int       loops  = 0;
                    
    	    if (syscall == NULL) { return -ENOSYS; }
            
            ASSERT( !(hmpv_read_esr_cl0()  & 0x6),("FAULT IN SYSCALL HANDLER (esr   = 0x%08x)!!", hmpv_read_esr_cl0()  ));
            ASSERT( !(hmpv_read_cesr_cl0() & 0xf),("FAULT IN SYSCALL HANDLER (cesr0 = 0x%08x)!!", hmpv_read_cesr_cl0() ));
            ASSERT( !(hmpv_read_cesr_cl1() & 0xf),("FAULT IN SYSCALL HANDLER (cesr1 = 0x%08x)!!", hmpv_read_cesr_cl1() ));
            
    	    save_and_cli(flags);
	    __ste();

            CHECK_STACK("C");

	    /* enable exceptions during syscalls */
	    __sti();

	    if (tsk->ptrace & PT_TRACESYS) {
		    syscall_trace();
	    }
            retval = syscall(a0, a1, a2, a3, a4, a5, a6, ctx);

           /*
            * We might be in a child process now if the
            * system call was a clone, or fork. 
            * Adjust the context pointer:
            */
            ctx= (void*) ( ((unsigned long)ctx & (KERNEL_STACK_SIZE - 1))
                         | ((unsigned long)current)
                         );
                       
           /*
            * In case the current system call is sigsuspend,
            * then the current context type may already be 
            * changed here to an INITIAL_CONTEXT to let the 
            * process proceed in user mode with the signal 
            * handler of the signal that awoke sigsuspend.
            *
            * In the other case the context type is still 
            * a SYSCALL_CONTEXT, but may be changed later on 
            * in case a signal is pending, by the below call \
            * to do_signal
            */
            if (ctx->header.u.s.type == ETI_SYSCALL_CONTEXT) {

                ctx->result = retval;

                if (tsk->ptrace & PT_TRACESYS) { syscall_trace(); }
            }

            RUN_BOTTOM;
            
           /***********************/
            do {
                DO_RESCHED(ctx);

                if ( user_mode(ctx) 
                  && current->sigpending 
                  && (ctx->header.u.s.type != ETI_INITIAL_CONTEXT)
                   ) {
           	    do_signal(0,ctx);
                }
                __cle();

		ASSERT( loops++ < 20, ("Looping in system call return loop!\n") );
            } while (
                (ctx->header.u.s.type != ETI_INITIAL_CONTEXT)
                     &&
                !return_address_present(ctx)
                    );
           /***********************/

            CHECK_STACK("c");
	    restore_flags(flags);

           /* 
            * In case context has been modified to handle a signal
            * then use the restore context routine; otherwise
            * take the quick return.
            */
            if (ctx->header.u.s.type == ETI_INITIAL_CONTEXT) {
                ctxRestoreInitialContext(ctx);
            } else {
                return ctx->result;
            }
    }   
}





void ALL_BREAKPOINT_EXCEPTIONS_handler(ctxExceptionContext_t ctx, UInt esr) ;
void CL_BREAKPOINT_EXCEPTIONS_handler(ctxExceptionContext_t ctx, UInt cesr, UInt cluster);
void ILPC_handler(ctxExceptionContext_t ctx, UInt esr) ;
void ALL_ILLEGAL_EXCEPTIONS_handler(ctxExceptionContext_t ctx, UInt cesr, UInt cluster) ;
void ALL_ITLB_EXCEPTIONS_handler(ctxExceptionContext_t ctx, UInt esr) ;
void ALL_DTLB_EXCEPTIONS_handler(ctxExceptionContext_t ctx, UInt cesr, UInt cluster) ;

void atomic_fixup( ctxExceptionContext_t ctx ) 
{
    struct task_struct *tsk = current;
    UInt32 epcr0 = ctx->epcr0;

    /* atomic operations: fixup if we've had an interrupt during
     * an atomic operation. */
    if (user_mode(ctx) && tsk->tss.atomic_start) {
        u32 mstat0     = ctx->cl[0].mstat0;
        u32 memtype    = mstat0 >> 28; /* JVDS MAGIC CONSTANT ALERT! */
        u32 iswrite    = (memtype&2);
        u32 isread     = (memtype&1);
        u32 isweird0   = (isread | iswrite) == 3;
        u32 isweird1   = (isread | iswrite) == 0;
        u32 isvalid    = (memtype&8) && !isweird0 && !isweird1;

        if ((epcr0 >= tsk->tss.atomic_start) &&
            (epcr0 <  tsk->tss.atomic_end)) {

            /* clear valid bit (so that load doesn't get issued since we're 
             *                  going to re-issue it anyways.) */
            if (isvalid)
                ctx->cl[0].mstat0 &= ~(1<<31);

            /* re-execute atomic op */
            ctx->epcr0 = tsk->tss.atomic_start;
            ctx->epcr1 = ctx->epcr2 = 0;
        } 

        else if (epcr0 == tsk->tss.atomic_end)
        {
            /* check if we had a DTLB miss on the store */
            if (isvalid && (ctx->cl[0].mvadr0 == ctx->cl[0].r32)) {

                /* clear valid bit */
                ctx->cl[0].mstat0 &= ~(1<<31); 

                /* re-execute atomic op */
                ctx->epcr0 = tsk->tss.atomic_start;
                ctx->epcr1 = ctx->epcr2 = 0;
            }

            /* otherwise, we simply had some other interrupt happen
             * after the store was completed.  Just continue on... */
        }
    }

}

static void ret_from_intr( ctxExceptionContext_t ctx )
{
    atomic_fixup(ctx);

    do {
        UInt32 exceptions, interrupts;

        bspregDisableInterrupts (interrupts); 
        bspregEnableExceptions  (exceptions);
        {
            DO_RESCHED(ctx);
            DO_SIGNAL(ctx);
        }
        bspregDisableExceptions (exceptions);
        bspregRestoreInterrupts (interrupts);
    } while (
          (ctx->header.u.s.type != ETI_INITIAL_CONTEXT)
               &&
          !return_address_present(ctx)
            );

    ASSERT (  !(hmpv_read_psw_cl0() & 2), 
             ("Exceptions enabled at end of handler\n") ); 
}











__initfunc(void trap_init(void)) 
{ 
    excSetup_t       setup;
    excExceptions_t *handle;


    excOpen( &handle );
    
   /* Get the default setup from libdev */ 
    excGetSetup(handle, &setup);

  /* Exceptions */
  /* ========== */
    setup.ILPC    = ILPC_handler;
    setup.IBPT    = ALL_BREAKPOINT_EXCEPTIONS_handler;
    setup.BPOP    = ALL_BREAKPOINT_EXCEPTIONS_handler;
    setup.ITLBAA  = ALL_ITLB_EXCEPTIONS_handler;
    setup.ITLBR   = ALL_ITLB_EXCEPTIONS_handler;
    setup.ITLBM   = ALL_ITLB_EXCEPTIONS_handler;

  /* Cluster Exceptions */
  /* ================== */
    setup.IEXZ    = ALL_ILLEGAL_EXCEPTIONS_handler;
    setup.ILLO    = ALL_ILLEGAL_EXCEPTIONS_handler;
    setup.PLV     = ALL_ILLEGAL_EXCEPTIONS_handler;
//    setup.DBPT  is left to the default one ;
    setup.DALN    = ALL_ILLEGAL_EXCEPTIONS_handler;
    setup.DTLBKW  = ALL_DTLB_EXCEPTIONS_handler;
    setup.DTLBAW  = ALL_DTLB_EXCEPTIONS_handler;
    setup.DTLBAA  = ALL_DTLB_EXCEPTIONS_handler;
    setup.DTLBR   = ALL_DTLB_EXCEPTIONS_handler;
    setup.DTLBM   = ALL_DTLB_EXCEPTIONS_handler;

  /* Syscall Handler */
  /* ====================== */
    setup.syscallHandler   = syscall_handler;

  /* Preprocessor function */
  /* ====================== */
    setup.preProcess       = Null;

  /* Postprocessor function */
  /* ====================== */
    setup.postProcess      = ret_from_intr;


    excSetup( handle, &setup );
}


