///
/// Copyright (c) Equator Technologies, Inc. Unpublished
///
/// Equator Technologies, Inc.
/// Proprietary & Confidential
///
/// This source code and the algorithms implemented therein constitute
/// confidential information and may comprise trade secrets of Equator
/// or its associates, and any use thereof is subject to the terms and
/// conditions of the Non-Disclosure Agreement pursuant to which this
/// source code was originally received.
///

/*
 * Timing Parameters:
 *     Baud Rate is 115200
 *        234.3750 (  234) clocks/bit when MAP Core is running at  27.0MHz
 *       2226.5625 ( 2226) clocks/bit when MAP Core is running at 256.5MHz
 *     Baud Rate is 9600
 *       2812.5000 ( 2811) clocks/bit when MAP Core is running at  27.0MHz
 *      26666.6666 (26667) clocks/bit when MAP Core is running at 256.5MHz
 *           
 */

/* RS232C BIT TIME IN CORE CLOCK CYCLES */
/* NOTE: below time shall be multiple of 3 to avoid compilation errors */
#define RS232C_BIT_TIME		2226

#define RS232C_BIT_COUNTER	(RS232C_BIT_TIME/3-8)

#define PIO_BASE 0xffc00000
/* IIC_DDC_SELECT control registers and values */
#define IIC_DDC_SELECT_ADDR     (PIO_BASE + 0x00068000)
#define IIC_DDC_SELECT_DATA_ON  0xff030fde
#define IIC_DDC_SELECT_DATA_OFF 0xff030bde

/* REGISTER ALLOCATION */
#define rs_send_byte		r32
#define led_addr		r33
#define led_data_on		r34
#define led_data_off		r35
#define rs232c_pulse_counter	r36
#define bit_check_mask		r37
#define rs_send_bit		r38
#define rs_write_value		r39
#define rs_send_pattern		r40
#define loop_var		r41
#define hexdump_arg		r42
#define hexdump_charcount	r43
#define hexdump_ascii0		r44
#define hexdump_r63		r45
#define check_flag		pr2

target arch(mapca);
text;
mark_begin;
mark_entry caller_saves, regs, atypes();
_iicdebug_putchar::

/* LOAD CONSTANTS INTO REGISTERS */
instr	cl0 ialu sethi.32 led_data_off,32hi@(IIC_DDC_SELECT_DATA_OFF)
	cl0 ifgalu sethi.32 led_data_on,32hi@(IIC_DDC_SELECT_DATA_ON);
instr	cl0 ialu dincr.32 led_data_off,32lo@(IIC_DDC_SELECT_DATA_OFF)
	cl0 ifgalu dincr.32 led_data_on,32lo@(IIC_DDC_SELECT_DATA_ON);
instr	cl0 ialu sethi.32 led_addr,32hi@(IIC_DDC_SELECT_ADDR)
	cl0 ifgalu sethi.32 rs232c_pulse_counter,32hi@(RS232C_BIT_COUNTER);
instr	cl0 ialu dincr.32 led_addr,32lo@(IIC_DDC_SELECT_ADDR)
	cl0 ifgalu dincr.32 rs232c_pulse_counter,32lo@(RS232C_BIT_COUNTER);

/* KEEP IIC_DDC_SELECT HIGH and TAKE ONLY 8-bit value from the argument */
instr	cl0 ialu ncst.32 led_data_on, led_addr, zero
	cl0 ifgalu mv.32 rs_write_value, 16@255
	cl1 ialu mnop 4;
instr	cl0 ialu band.32 rs_send_byte, rs_send_byte, rs_write_value
	cl1 ialu mnop 4;

/* SEND BYTE IN RS-232C FORMAT */
instr	cl0 ialu sethi.32 bit_check_mask, 32hi@(0x00010000)
	cl0 ifgalu sethi.32 rs_write_value, 32hi@(0xfc010000);
instr   cl0 ifgalu ls.32 rs_send_pattern, rs_send_byte, 6@18
        cl1 ialu mnop 4;
instr   cl0 ialu bor.32 rs_send_pattern, rs_send_pattern, rs_write_value;

/* SEND BITS TO IIC_DDC_SELECT PIN */
rs_send_bit_loop:
instr	cl0 ialu band.32 rs_send_bit, rs_send_pattern, bit_check_mask;
instr	cl0 ialu cne.s32 check_flag, rs_send_bit, bit_check_mask
	cl0 ifgalu ls.32 bit_check_mask, bit_check_mask, 6@1
	cl1 ialu mnop 4;
instr   cl0 ialu slct.32 rs_write_value, check_flag, led_data_off, led_data_on
	cl1 ialu mnop 1;
instr	cl0 ialu ncst.32 rs_write_value, led_addr, zero
	cl0 ifgalu mv.32 loop_var, rs232c_pulse_counter
	cl1 ialu mnop 1
	cl1 ifgalu mv.1 pball.check_flag, pr1; 

rs_send_bit_wait_loop:
instr	cl0 ialu if check_flag bra 19@rs_send_bit_wait_loop; 
instr	cl0 ialu cgt.s32 check_flag, loop_var, zero;
instr	cl0 ialu sub.32 loop_var, loop_var, 6@1;

instr	cl0 ifgalu cne.s32 check_flag, bit_check_mask, zero
	cl1 ialu mnop 4;

instr	cl0 ialu if check_flag bra 19@rs_send_bit_loop;
instr;
instr;

/* RETURN */
instr	cl0 ialu bra r63;
instr;
instr;

_iicdebug_int32dump ::
/* assumes r32 to be rs_send_byte */
instr;
instr;
instr	cl0 ialu mv.32 hexdump_r63, r63
	cl0 ifgalu mv.32 hexdump_ascii0, 16@0x30;
instr	cl0 ialu mv.32 hexdump_arg, r33
	cl0 ifgalu add.32 hexdump_charcount, zero, 6@9;

hexdump_loop:
instr	cl0 ialu bra 19@_iicdebug_putchar;
instr;
instr	cl1 ialu pcadd.32 gb0.r63, 6@0;

instr	cl0 ifgalu rol.32 hexdump_arg, hexdump_arg, 6@4
	cl1 ialu mnop 5;
instr	cl0 ialu band.32 rs_send_byte, hexdump_arg, 6@15;
instr	cl0 ifgalu cgt.s32 check_flag, rs_send_byte, 6@9
	cl0 ialu sub.32 hexdump_charcount, hexdump_charcount, 6@1
	cl1 ialu mnop 5;
instr	cl0 ialu if check_flag add.32 rs_send_byte, rs_send_byte, 6@7
	cl1 ialu mnop 5;
instr	cl0 ialu add.32 rs_send_byte, rs_send_byte, hexdump_ascii0;

instr	cl0 ialu mnop 5;
instr	cl0 ialu mv.32 r63, hexdump_r63;
instr	cl0 ifgalu cgt.s32 check_flag, hexdump_charcount, 6@0
	cl1 ialu mnop 6;
instr	cl0 ialu if check_flag bra 19@hexdump_loop;
instr;
instr;

instr	cl0 ialu bra r63;
instr;
instr;

_iicdebug_end::

end;
