/*
 * elf.h
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) Equator Technologies
 */
#ifndef _ASM_BSP_ELF_H_
#define _ASM_BSP_ELF_H_

/* ETI ELF magic numbers */
#define EM_ETI_BSP1x   0xe817

#define ELF_NCLUSTERS 2
#define ELF_NGREGS (ELF_NCLUSTERS * 64)

typedef unsigned long elf_greg_t;
typedef elf_greg_t elf_gregset_t[ELF_NGREGS];

/* We don't have FP regs, but fs/binfmt_elf.c needs this type */
typedef double elf_fpreg_t;
typedef elf_fpreg_t elf_fpregset_t[0];

/*
 * This is used to ensure we don't load something for the wrong architecture.
 */
#define elf_check_arch(x)      ((x) == EM_ETI_BSP1x)
#define ELF_ARCH                EM_ETI_BSP1x
#define ELF_CLASS               ELFCLASS32
#define ELF_DATA                ELFDATA2LSB
#define USE_ELF_CORE_DUMP
#define ELF_EXEC_PAGESIZE	(1<<14)

#define ELF_CORE_COPY_REGS(_dest,_regs)				\
	memcpy((char *) &_dest, (char *) _regs,			\
	       sizeof(pt_regs));

/* This yields a mask that user programs can use to figure out what
   instruction set this cpu supports.  This could be done in userspace,
   but it's not easy, and we've already done it here.  */

#define ELF_HWCAP       (0)

/* This yields a string that ld.so will use to load implementation
   specific libraries for optimization.  This is more specific in
   intent than poking at uname or /proc/cpuinfo.

   For the moment, we have only optimizations for the Intel generations,
   but that could change... */

#define ELF_PLATFORM  (NULL)

/* This is the location that an ET_DYN program is loaded if exec'ed.  Typical
   use of this is to invoke "./ld.so someprog" to test out a new version of
   the loader.  We need to make sure that it is out of the way of the program
   that it will "exec", and that there is sufficient room for the brk.  */

#define ELF_ET_DYN_BASE         (2 * TASK_SIZE / 3)

#ifdef __KERNEL__
#define SET_PERSONALITY(ex,ibcs2) \
	current->personality = (ibcs2 ? PER_SVR4 : PER_LINUX)
#endif

#endif	/* _ASM_BSP_ELF_H_ */
