/*
 * io.h
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) Equator Technologies
 */

#ifndef _ASM_BSP_IO_H_
#define _ASM_BSP_IO_H_

#include <linux/config.h>

#include <asm/page_offset.h>
#include <asm/processor.h>
#include <asm/system.h>
#include <asm/types.h>
#include <asm/pci.h>

#include <libdev/etiPci.h>


extern inline void* phys_to_virt(unsigned long phys_addr)
{
    return (void*)(phys_addr + PAGE_OFFSET);
}

extern inline unsigned long virt_to_phys(void* virt_addr)
{
    return ((unsigned long)virt_addr) - PAGE_OFFSET;
}

extern inline unsigned long phys_to_bus(unsigned long phys_addr)
{
	UInt32 bar0 = PCI_PIO(Fn0BAR0);
	return phys_addr + (bar0 & PCI_BASE_ADDRESS_MEM_MASK);
}    

extern inline unsigned long bus_to_phys(unsigned long bus_addr)
{
	UInt32 bar0 = PCI_PIO(Fn0BAR0);
	return bus_addr - (bar0 & PCI_BASE_ADDRESS_MEM_MASK);
}

extern inline unsigned long virt_to_bus(void * virt_addr)
{
	return phys_to_bus( virt_to_phys(virt_addr) );
}

extern inline void * bus_to_virt(unsigned long bus_addr)
{
	return phys_to_virt( bus_to_phys(bus_addr) );
}


extern void *__pioremap(unsigned long phys_addr, unsigned long size, unsigned long flags);
extern void * __ioremap(unsigned long phys_addr, unsigned long size, unsigned long flags);
extern void *   ioremap(unsigned long phys_addr, unsigned long size);
extern void     iounmap(void *virt_addr);



/* The IO areas are already mapped non-cachable by the TLB settings */
#define ioremap_nocache ioremap

#define inb(port)	({ __u8 val;  pciReadIoUInt8 (port, &val); val; })
#define inw(port)	({ __u16 val; pciReadIoUInt16(port, &val); val; })
#define inl(port)	({ __u32 val; pciReadIoUInt32(port, &val); val; })

#define outb(data, port)	pciWriteIoUInt8 (port, data)
#define outw(data, port)	pciWriteIoUInt16(port, data)
#define outl(data, port)	pciWriteIoUInt32(port, data)

#define insb(port, dst, count)	pciReadIoUInt8S (port, count, dst)
#define insw(port, dst, count)	pciReadIoUInt16S(port, count, dst)
#define insl(port, dst, count)	pciReadIoUInt32S(port, count, dst)

#define outsb(port, src, count)	pciWriteIoUInt8S (port, count, src)
#define outsw(port, src, count)	pciWriteIoUInt16S(port, count, src)
#define outsl(port, src, count)	pciWriteIoUInt32S(port, count, src)

#define outb_p	outb
#define outw_p	outw
#define outl_p	outl
#define inb_p	inb
#define inw_p	inw
#define inl_p	inl

/*
 * /writeX() are used to access memory mapped devices. On some
 * architectures the memory mapped IO stuff needs to be accessed
 * differently. On the BSP architecture, we just read/write the
 * memory location directly, but using uncached loads/stores to ensure
 * strong ordering.
 */

#define readb(addr)	(*(volatile __uncached__ unsigned char  *)(addr))
#define readw(addr)	(*(volatile __uncached__ unsigned short *)(addr))
#define readl(addr)	(*(volatile __uncached__ unsigned int   *)(addr))

#define writeb(b,addr)	(*(volatile __uncached__ unsigned char  *)(addr) = (b))
#define writew(b,addr)	(*(volatile __uncached__ unsigned short *)(addr) = (b))
#define writel(b,addr)	(*(volatile __uncached__ unsigned int   *)(addr) = (b))

#define memset_io(a,b,c)	memset((a),(b),(c))
#define memcpy_fromio(a,b,c)	memcpy((a),(b),(c))
#define memcpy_toio(a,b,c)	memcpy((a),(b),(c))

#endif	/* _ASM_BSP_IO_H_ */
